pfUI:RegisterModule("lootfilter", function ()

  local msg_pref = GetAddOnMetadata("pfUI-lootfilter", "Title") .. "|r: "
  local min_item_id = 1
  local max_item_id = 25818
  local _list_names = {"vendor", "delete"}

  -- local config = C.lootfilter
  local config

  local function GetItemIdByName(name)
    local found_item_id
    for item_id = min_item_id, max_item_id do
      local item_name, _, _ = GetItemInfo(item_id)
      if (item_name and strupper(item_name) == strupper(name)) then
        found_item_id = tostring(item_id)
        break
      end
    end
    return found_item_id
  end

  -- migrate config to item_id from item_name
  local function MigrateListItemID()
    for list_index in _list_names do
      local temp_str = ""
      if config[_list_names[list_index]].list then
        for _, list_item in pairs({strsplit("#", config[_list_names[list_index]].list)}) do
          if tonumber(list_item) then
            if (tonumber(list_item) >= min_item_id and tonumber(list_item) <= max_item_id) then
              temp_str = temp_str .. "#" .. list_item
            end
          else
            local item_id = GetItemIdByName(list_item)
            if item_id then
              temp_str = temp_str .. "#" .. item_id
            end
          end
        end
        if not config[_list_names[list_index]].list_ids then
          config[_list_names[list_index]].list_ids = temp_str
        else
          config[_list_names[list_index]].list_ids =  config[_list_names[list_index]].list_ids .. temp_str
        end
        config[_list_names[list_index]].list = nil
      end
    end
  end

  local delete_que = {}
  local vendor_que = {}

  pfUI.lootfilter = CreateFrame("Frame", "pfLootfilter", UIParent)
  pfUI.lootfilter:RegisterEvent("VARIABLES_LOADED")
  pfUI.lootfilter.cache = {}
  pfUI.lootfilter.active = false

  local function GenCache()
    for list_index in _list_names do
      local tmp_display = ""
      pfUI.lootfilter.cache[_list_names[list_index]] = {}
      pfUI.lootfilter.cache[_list_names[list_index]].list = {}
      pfUI.lootfilter.cache[_list_names[list_index]].str_display = ""
      for _, item_id in pairs({strsplit("#", config[_list_names[list_index]].list_ids)}) do
        pfUI.lootfilter.cache[_list_names[list_index]].list[item_id] = ""
        local item_name, _, _ = GetItemInfo(item_id)
        tmp_display = tmp_display .. "#" .. item_id .. "::" .. item_name
      end
      pfUI.lootfilter.cache[_list_names[list_index]].str_display = tmp_display
    end
  end


  pfUI.lootfilter.autodelete = CreateFrame("Frame", "pfLootfilterAutodelete", pfUI.lootfilter)
  pfUI.lootfilter.autodelete:RegisterEvent("BAG_UPDATE")  
  pfUI.lootfilter.autodelete:RegisterEvent("UNIT_INVENTORY_CHANGED")

  pfUI.lootfilter.autovendor = CreateFrame("Frame", "pfLootfilterAutovendor", pfUI.lootfilter)
  pfUI.lootfilter.autovendor:RegisterEvent("MERCHANT_SHOW")
  pfUI.lootfilter.autovendor:RegisterEvent("MERCHANT_CLOSED")
  pfUI.lootfilter.autovendor:Hide()

  local function TableMatch(table, needle)
    for i, v in ipairs(table) do
      if strupper(v) == strupper(needle) then
        return i
      end
    end
    return false
  end

  local function ListStringRemove(istr, needle)
    local newstr = ""
    for i, val in pairs({strsplit("#", istr)}) do
      if val ~= needle then
        newstr = newstr .. "#" .. val
      end
    end
    return newstr
  end

  local function ScanDeleteItems()
    delete_que = {}
    for bag = 0, 4, 1 do
      for slot = 1, GetContainerNumSlots(bag), 1 do
        local item_link = GetContainerItemLink(bag, slot)
        if item_link then
          local _, _, item_id = strfind(item_link, "(%d+):")
          if item_id and pfUI.lootfilter.cache.delete.list[item_id] then
            table.insert(delete_que, {bag, slot})
          end
        end
      end
    end
  end

  local function GetNextDeleteItem()
    if delete_que[1] then
      local bag, slot = delete_que[1][1], delete_que[1][2]
      table.remove(delete_que, 1)
      return bag, slot
    else
      return nil, nil
    end
  end

  local function ProcessDeleteItems()
    local bag, slot = GetNextDeleteItem()
    if not bag or not slot then return end

    local item_link = GetContainerItemLink(bag, slot)
    if item_link then
      local _, _, item_id = strfind(item_link, "(%d+):")
      if item_id and pfUI.lootfilter.cache.delete.list[item_id] then
        PickupContainerItem(bag, slot)
        DeleteCursorItem()
        if config.delete.notification == "1" then
          local item_name, _, _ = GetItemInfo(item_id)
          local item_link = pfUI.api.GetItemLinkByName(item_name)
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "deleted " .. item_link)
        end
        return
      end
    end
  end

  pfUI.lootfilter.autodelete:SetScript("OnEvent", function()
    if not pfUI.lootfilter.active then return end
    if event == "UNIT_INVENTORY_CHANGED" and arg1 ~= "player" then return end
    pfUI.api.QueueFunction(ScanDeleteItems)
  end)

  pfUI.lootfilter.autodelete:SetScript("OnUpdate", function()
    if not pfUI.lootfilter.active then return end
    if (this.tick or 1) > GetTime() then return else this.tick = GetTime() + 0.3 end
    pfUI.api.QueueFunction(ProcessDeleteItems)
  end)

  local function ScanVendorItems()
    vendor_que = {}
    for bag = 0, 4, 1 do
      for slot = 1, GetContainerNumSlots(bag), 1 do
        local item_link = GetContainerItemLink(bag, slot)
        if item_link then
          local _, _, item_id = strfind(item_link, "(%d+):")
          if item_id and pfUI.lootfilter.cache.vendor.list[item_id] then
            table.insert(vendor_que, {bag, slot})
          end
        end
      end
    end
  end

  local function GetNextVendorItem()
    if vendor_que[1] then
      local bag, slot = vendor_que[1][1], vendor_que[1][2]
      table.remove(vendor_que, 1)
      return bag, slot
    else
      return nil, nil
    end
  end

  pfUI.lootfilter.autovendor:SetScript("OnShow", function()
    this.sell_count = 0
    this.sell_price = 0
  end)

  pfUI.lootfilter.autovendor:SetScript("OnHide", function()
    if this.sell_count > 0 and config.vendor.notification == "1" then
      DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "items has been sold and you earned" .. " " .. CreateGoldString(this.sell_price))
    end
  end)

  pfUI.lootfilter.autovendor:SetScript("OnEvent", function()
    if pfUI.lootfilter.active then
      if event == "MERCHANT_SHOW" then
        ScanVendorItems()
        this:Show()
      elseif event == "MERCHANT_CLOSED" then
        this:Hide()
      end
    end
  end)

  pfUI.lootfilter.autovendor:SetScript("OnUpdate", function()
    if (this.tick or 0.2) > GetTime() then return else this.tick = GetTime() + 0.2 end
    local bag, slot = GetNextVendorItem()
    if bag == nil or slot == nil then
      this:Hide()
      return
    end

    local item_link = GetContainerItemLink(bag, slot)
    if item_link then
      local _, _, item_id = strfind(item_link, "(%d+):")
      if item_id and pfUI.lootfilter.cache.vendor.list[item_id] then
        local _, stack_count = GetContainerItemInfo(bag, slot)
        local item_name, _, _ = GetItemInfo(item_id)
        if pfSellData[tonumber(item_id)] then
          local _, _, sell, buy = strfind(pfSellData[tonumber(item_id)], "(.*),(.*)")
          this.sell_price = this.sell_price + (sell * (tonumber(stack_count) or 1))
          this.sell_count = this.sell_count + 1
        end
        UseContainerItem(bag, slot)
      end
    end
  end)

  local function UpdateLists()
    for list_index in _list_names do
      local tmp_str = ""
      for item_id, _ in pfUI.lootfilter.cache[_list_names[list_index]].list do
        tmp_str = tmp_str .. "#" .. item_id
      end
      config[_list_names[list_index]].list_ids = tmp_str
    end
  end

  local function UpdateCache()
    for list_index in _list_names do
      local tmp_str = ""
      local update_table = {}
      for i, item_string in pairs({strsplit("#", pfUI.lootfilter.cache[_list_names[list_index]].str_display)}) do
        local key, value = strsplit("::", item_string)
        if key then
          local item_id = key
          if not tonumber(key) then
            item_id = GetItemIdByName(key)
          end
          if item_id then
            if (tonumber(item_id) >= min_item_id and tonumber(item_id) <= max_item_id) then
              if update_table[item_id] then
                update_table[item_id] = nil
              else
                update_table[item_id] = ""
              end
            end
          end
        end
      end
      for item_id in update_table do
        local item_name, _, _ = GetItemInfo(item_id)
        tmp_str = tmp_str .. "#" .. item_id .. "::" .. item_name
      end
      pfUI.lootfilter.cache[_list_names[list_index]].str_display = tmp_str
      for item_id, _ in pfUI.lootfilter.cache[_list_names[list_index]].list do
        if not update_table[item_id] then
          pfUI.lootfilter.cache[_list_names[list_index]].list[item_id] = nil
        end
      end
      for item_id, _ in update_table do
        if not pfUI.lootfilter.cache[_list_names[list_index]].list[item_id] then
          pfUI.lootfilter.cache[_list_names[list_index]].list[item_id] = ""
        end
      end
    end
    UpdateLists()
  end

  function pfUI.lootfilter:Load()
    config = C.lootfilter
    pfUI.lootfilter:UnregisterEvent("VARIABLES_LOADED")
    pfUI.lootfilter:RegisterEvent("PLAYER_LOGIN")
    pfUI.lootfilter:SetScript("OnEvent", GenCache)

      -- migrate old config entires
    if config.delete_filter ~= nil then
      local temp_del_str = ""
      if type(config.delete_filter) == "table" then
        for i, item_name in pairs(config.delete_filter) do
          temp_del_str = temp_del_str .. "#" .. item_name
        end
        config.delete.list = config.delete.list .. temp_del_str
        config.delete_filter = nil
      elseif type(config.delete_filter) == "string" then
        for i, item_name in pairs({strsplit("#", config.delete_filter)}) do
          temp_del_str = temp_del_str .. "#" .. item_name
        end
        config.delete.list = config.delete.list .. temp_del_str
        config.delete_filter = nil
      else
        config.delete_filter = nil
      end
    end

    if config.vendor_filter ~= nil then
      local temp_del_str = ""
      if type(config.vendor_filter) == "table" then
        for i, item_name in pairs(config.vendor_filter) do
          temp_del_str = temp_del_str .. "#" .. item_name
        end
        config.vendor.list = config.vendor.list .. temp_del_str
        config.vendor_filter = nil
      elseif type(config.vendor_filter) == "string" then
        for i, item_name in pairs({strsplit("#", config.vendor_filter)}) do
          temp_del_str = temp_del_str .. "#" .. item_name
        end
        config.vendor.list = config.vendor.list .. temp_del_str
        config.vendor_filter = nil
      else
        config.vendor_filter = nil
      end
    end

    -- query server for missing items in cache
    for item_id = min_item_id, max_item_id do
      if not GetItemInfo(item_id) then
        GameTooltip:SetHyperlink("item:" .. item_id .. ":0:0:0")
        GameTooltip:Hide()
      end
    end

    MigrateListItemID()

  end

  function pfUI.lootfilter:UpdateConfig()
    UpdateCache()
    pfUI.lootfilter.autodelete:GetScript("OnEvent")()
  end

  pfUI.lootfilter:SetScript("OnEvent", pfUI.lootfilter.Load)

  local function SlashHandler(msg)
    local _, _, cmd, arg = string.find(msg, "%s?(%w+)%s?(.*)")

    local function HelpCommand()
      DEFAULT_CHAT_FRAME:AddMessage(msg_pref)
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf help|r - " .. "Print out all available commands")
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf toggle|r - " .. "Turn filtering ON/OFF (OFF by default)")
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf vendor <LINK | NAME | ITEM_ID>|r - " .. "Add/Remove item to vendor list by LINK or NAME or ITEM_ID")
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf delete <LINK | NAME | ITEM_ID>|r - " .. "Add/Remove item to delete list by LINK or NAME or ITEM_ID")
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf export <LIST_NAME>|r - " .. "Export LIST_NAME (vendor, delete) data in a format LIST_NAME:::DATA")
      DEFAULT_CHAT_FRAME:AddMessage("|cff33ffcc/pflf import <IMPORT_STRING>|r - " .. "Import IMPORT_STRING data of a format LIST_NAME:::DATA")
    end

    local function AddRemove(list, item)
      local item_id
      if tonumber(item) then
        item_id = tostring(item)
      else
        if GetItemIdByName(item) then
          item_id = GetItemIdByName(item)
        else
          local _, _, _item_id = strfind(item, "(%d+):")
          item_id = _item_id
        end
      end
      if (tonumber(item_id) and (tonumber(item_id) >= min_item_id and tonumber(item_id) <= max_item_id)) then
        local item_name, _, _ = GetItemInfo(item_id)
        local item_link = pfUI.api.GetItemLinkByName(item_name)
        pfUI.lootfilter.cache[list].str_display = pfUI.lootfilter.cache[list].str_display .. "#" .. item_id .. "::" .. item_name

        if pfUI.lootfilter.cache[list].list[item_id] then
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. item_link .. " removed from " .. list .. " filter list")
        else
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. item_link .. " added to " .. list .. " filter list")
        end
        UpdateCache()
      end
    end

    local function ExportList(list)
      if config[list].list_ids then
        if config[list].list_ids == "" then
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "list is empty")
        else
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "export string\n" .. list .. ":::" .. config[list].list_ids)
        end
      else
        DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "list not found, use |cff33ffcc/pflf help|r")
      end
    end

    local function ImportList(list_string)
      local list, data = strsplit(":::", list_string)
      if list and data then
        if pfUI.lootfilter.cache[list] then
          pfUI.lootfilter.cache[list].str_display = data
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. list .. " list imported")
          UpdateCache()
        end
      else
        DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "list not found or no data provided, use |cff33ffcc/pflf help|r")
      end
    end
    
    local commands = {
      ["help"] = function() HelpCommand() end,
      ["toggle"] = function()
        pfUI.lootfilter.active = not pfUI.lootfilter.active
        if pfUI.lootfilter.active then
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "enabled")
        else
          DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "disabled")
        end
      end,
      ["vendor"] = function() AddRemove("vendor", arg) end,
      ["ven"] = function() AddRemove("vendor", arg) end,
      ["delete"] = function() AddRemove("delete", arg) end,
      ["del"] = function() AddRemove("delete", arg) end,
      ["export"] = function() ExportList(arg) end,
      ["import"] = function() ImportList(arg) end,
    }

    if not cmd or cmd == "" or cmd == "help" then
      commands["help"]()
    elseif not commands[cmd] then
      DEFAULT_CHAT_FRAME:AddMessage(msg_pref .. "command not found, use |cff33ffcc/pflf help|r")
    else
      commands[cmd]()
    end
    return
  end

  _G.SLASH_PFLF1 = "/pflf"
  _G.SlashCmdList.PFLF = SlashHandler

  pfUI.lootfilter.C = config

end)